/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package age

import (
	"bytes"
	"fmt"
)

type AgeError struct {
	cause error
	msg   string
}

func (e *AgeError) Error() string {
	if e.cause != nil {
		return fmt.Sprintf("%s >> Cause:%s", e.msg, e.cause.Error())
	}
	return e.msg
}

type AgeParseError struct {
	msg    string
	errors []string
}

func (e *AgeParseError) Error() string {
	var buf bytes.Buffer
	buf.WriteString(e.msg)
	buf.WriteString(" >> Causes:\n")
	for _, err := range e.errors {
		buf.WriteString(err)
		buf.WriteString("\n")
	}
	return buf.String()
}
