PG_KEYWORD("all", ALL, RESERVED_KEYWORD)
PG_KEYWORD("analyze", ANALYZE, RESERVED_KEYWORD)
PG_KEYWORD("and", AND, RESERVED_KEYWORD)
PG_KEYWORD("as", AS, RESERVED_KEYWORD)
PG_KEYWORD("asc", ASC, RESERVED_KEYWORD)
PG_KEYWORD("ascending", ASCENDING, RESERVED_KEYWORD)
PG_KEYWORD("by", BY, RESERVED_KEYWORD)
PG_KEYWORD("call", CALL, RESERVED_KEYWORD)
PG_KEYWORD("case", CASE, RESERVED_KEYWORD)
PG_KEYWORD("coalesce", COALESCE, RESERVED_KEYWORD)
PG_KEYWORD("contains", CONTAINS, RESERVED_KEYWORD)
PG_KEYWORD("create", CREATE, RESERVED_KEYWORD)
PG_KEYWORD("delete", DELETE, RESERVED_KEYWORD)
PG_KEYWORD("desc", DESC, RESERVED_KEYWORD)
PG_KEYWORD("descending", DESCENDING, RESERVED_KEYWORD)
PG_KEYWORD("detach", DETACH, RESERVED_KEYWORD)
PG_KEYWORD("distinct", DISTINCT, RESERVED_KEYWORD)
PG_KEYWORD("else", ELSE, RESERVED_KEYWORD)
PG_KEYWORD("end", END_P, RESERVED_KEYWORD)
PG_KEYWORD("ends", ENDS, RESERVED_KEYWORD)
PG_KEYWORD("exists", EXISTS, RESERVED_KEYWORD)
PG_KEYWORD("explain", EXPLAIN, RESERVED_KEYWORD)
PG_KEYWORD("false", FALSE_P, RESERVED_KEYWORD)
PG_KEYWORD("in", IN, RESERVED_KEYWORD)
PG_KEYWORD("is", IS, RESERVED_KEYWORD)
PG_KEYWORD("limit", LIMIT, RESERVED_KEYWORD)
PG_KEYWORD("match", MATCH, RESERVED_KEYWORD)
PG_KEYWORD("merge", MERGE, RESERVED_KEYWORD)
PG_KEYWORD("not", NOT, RESERVED_KEYWORD)
PG_KEYWORD("null", NULL_P, RESERVED_KEYWORD)
PG_KEYWORD("optional", OPTIONAL, RESERVED_KEYWORD)
PG_KEYWORD("or", OR, RESERVED_KEYWORD)
PG_KEYWORD("order", ORDER, RESERVED_KEYWORD)
PG_KEYWORD("remove", REMOVE, RESERVED_KEYWORD)
PG_KEYWORD("return", RETURN, RESERVED_KEYWORD)
PG_KEYWORD("set", SET, RESERVED_KEYWORD)
PG_KEYWORD("skip", SKIP, RESERVED_KEYWORD)
PG_KEYWORD("starts", STARTS, RESERVED_KEYWORD)
PG_KEYWORD("then", THEN, RESERVED_KEYWORD)
PG_KEYWORD("true", TRUE_P, RESERVED_KEYWORD)
PG_KEYWORD("union", UNION, RESERVED_KEYWORD)
PG_KEYWORD("unwind", UNWIND, RESERVED_KEYWORD)
PG_KEYWORD("verbose", VERBOSE, RESERVED_KEYWORD)
PG_KEYWORD("when", WHEN, RESERVED_KEYWORD)
PG_KEYWORD("where", WHERE, RESERVED_KEYWORD)
PG_KEYWORD("with", WITH, RESERVED_KEYWORD)
PG_KEYWORD("xor", XOR, RESERVED_KEYWORD)
PG_KEYWORD("yield", YIELD, RESERVED_KEYWORD)