/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

LOAD 'age';
SET search_path TO ag_catalog;

-- basic comparisons
SELECT '0'::graphid = '0'::graphid, '0'::graphid = '1'::graphid;
SELECT '0'::graphid <> '0'::graphid, '0'::graphid <> '1'::graphid;
SELECT '0'::graphid < '1'::graphid,
       '0'::graphid < '0'::graphid,
       '1'::graphid < '0'::graphid;
SELECT '0'::graphid > '1'::graphid,
       '0'::graphid > '0'::graphid,
       '1'::graphid > '0'::graphid;
SELECT '0'::graphid <= '1'::graphid,
       '0'::graphid <= '0'::graphid,
       '1'::graphid <= '0'::graphid;
SELECT '0'::graphid >= '1'::graphid,
       '0'::graphid >= '0'::graphid,
       '1'::graphid >= '0'::graphid;

-- b-tree index
CREATE TABLE graphid_table (gid graphid);
INSERT INTO graphid_table VALUES ('0'), ('1'), ('2');
CREATE INDEX ON graphid_table (gid);
SET enable_seqscan = OFF;
EXPLAIN (COSTS FALSE) SELECT * FROM graphid_table WHERE gid = '1';
EXPLAIN (COSTS FALSE) SELECT * FROM graphid_table WHERE gid > '0';
SET enable_seqscan = ON;
DROP TABLE graphid_table;
